% interpolateFEAresults.m
% Import Finite Element Analysis (FEA) nodal results
% Import Universal File Format (UFF) or UNV file data from I-DEAS
% 2005-09-24, D. Coyne
% Notes:
% 1) Using an extension of the UFF Read m-file, readuff.m, dated 9/24/2005
% 2) This version is written for the results of a specific ITM FEA;
%    adaptation/generalization for another finite element model requires
%    some editing.

[UffDataSets,Info,errmsg] = readuff('ITM_gravity_sag.unv');

if Info.nErrors ~= 0
    for ii=1:Info.nErrors
        disp(Info.errorMsgs{ii});
    end
end

% echo header information
iHeader=find(Info.dsTypes==151);
UffDataSets{iHeader}

% echo units information
iUnits=find(Info.dsTypes==164);
UffDataSets{iUnits}

% nodal information
iNodal=find(Info.dsTypes==2411);
UffDataSets{iNodal}

% permanent groups
iGroups=find(Info.dsTypes==2452);
UffDataSets{iGroups}
% in this case the dataset I want is the the 8th
% labeled "front surface nodes"
UffDataSets{iGroups}.groupName{8}
frontSurfaceNodes = UffDataSets{iGroups}.entityTag{8};
nFrontSuraceNodes = size(frontSurfaceNodes,2);

% nodal analysis information
iNodalResults=find(Info.dsTypes==2414);
UffDataSets{iNodalResults}
% in this case the first dataset has the displacements
% and is the one I want
nNodalSets = length(iNodalResults);

% node numbers
nodeNumbers = UffDataSets{iNodalResults(1)}.nodeNum;
nNodes = length(nodeNumbers);

% extract front surface node group positions
for i=1:nFrontSuraceNodes
   pos(i)=find(UffDataSets{iNodal}.nodeLabel==frontSurfaceNodes(i));
   frontSurfaceNodePositions(i,:)=[UffDataSets{iNodal}.x(pos(i)),UffDataSets{iNodal}.y(pos(i)),UffDataSets{iNodal}.z(pos(i))];
end

% Plot Front Surface Nodal Pattern
figure(1)
plot3(frontSurfaceNodePositions(:,1),frontSurfaceNodePositions(:,2),frontSurfaceNodePositions(:,3),'go')
axis equal
title('Surface Nodal Pattern')
%print surface_nodes.ps -dpsc2

% extract front surface node group deformation
for i=1:nFrontSuraceNodes
   pos(i)=find(UffDataSets{iNodalResults(1)}.nodeNum==frontSurfaceNodes(i));
   frontSurfaceNodeDeform(i,:)=[UffDataSets{iNodalResults(1)}.r1(pos(i)),UffDataSets{iNodalResults(1)}.r2(pos(i)),UffDataSets{iNodalResults(1)}.r3(pos(i))];
end

% Plot Front Surface Deformation Shape
scale=0.05;
figure(2)
plot3(frontSurfaceNodePositions(:,1)+scale*frontSurfaceNodeDeform(:,1), ...
    frontSurfaceNodePositions(:,2)+scale*frontSurfaceNodeDeform(:,2), ...
    frontSurfaceNodePositions(:,3)+scale*frontSurfaceNodeDeform(:,3),'go')
axis square
grid on
title('Surface Deformation')
rotate3d on

% Interpolate from nonuniform nodal grid to a uniform grid & plot
	ngrid=51;
    w = 0.20;
	dx = w/(ngrid-1);
	xi = -w/2:dx:w/2;
	yi=xi';
	[xi,yi,zi]=griddata(frontSurfaceNodePositions(:,1),frontSurfaceNodePositions(:,2),frontSurfaceNodeDeform(:,3),xi,yi);
	figure;
	surfc(xi,yi,zi)
	shading interp
	title(strcat('Front Surface Displacement from Gravity Load'))
    xlabel('Horizontal Distance from Center (m)');
    ylabel('Vertical Distance from Center (m)');
    zlabel('Deformation Normal to Surface (m)')
	rotate3d on
   
    figure;
    v = -2e-9:0.1e-9:2e-9;
    [C,h] = contour(xi,yi,zi,v);
    axis equal;
    clabel(C,h)
    xlabel('Horizontal Distance from Center (m)');
    ylabel('Vertical Distance from Center (m)');
   
% least squares fit to Zernikes
    xf=reshape(xi,1,ngrid^2);
    yf=reshape(yi,1,ngrid^2);
    zf=reshape(zi,1,ngrid^2);
    rad = sqrt(xf.^2 + yf.^2);
    theta = atan2(yf, xf);
    % choose only points in the central region !!!!
    radiusCentral = 0.06;
    disp(['central region radius = ',num2str(radiusCentral,3),' (m)']);
    include=find(rad <= radiusCentral);
    radiusNormalize = min(radiusCentral,max(rad));
    r = rad(include)/radiusNormalize;
    t = theta(include);
    z = zi(include);
    
    % calculate the mean and rms in the central region
    zMean = mean(z);
    zPV = max(z) - min(z);
    zRms = norm(z)/sqrt(length(z));
    disp(['mean surface deformation in the central region = ',num2str(zMean,5),' (m)'])
    disp(['p-v surface deformation in the central region = ',num2str(zPV,5),' (m)'])
    disp(['rms surface deformation in the central region = ',num2str(zRms,5),' (m)'])
    
    % Zernike decomposition
    A = [ones(size(r)); r.*cos(t); r.*sin(t); -1+2*r.^2; r.^2.*cos(2*t); r.^2.*sin(2*t); ...
            (3*r.^2 - 2).*r.*cos(t); (3*r.^2 - 2).*r.*sin(t); 6*r.^4 - 6*r.^2 + 1]';
    Ainv = pinv(A);
    zernf = z*Ainv';
    disp('Zernike fits (nm):')
    disp(['Z1, piston           = ',num2str(zernf(1)*1e9,5)])
    disp(['Z2, yaw              = ',num2str(zernf(2)*1e9,5)])
    disp(['Z3, pitch            = ',num2str(zernf(3)*1e9,5)])
    disp(['Z4, focus            = ',num2str(zernf(4)*1e9,5)])
    disp(['Z5, astigmatism (0)  = ',num2str(zernf(5)*1e9,5)])
    disp(['Z6, astigmatism (45) = ',num2str(zernf(6)*1e9,5)])
    disp(['Z7, x-Coma           = ',num2str(zernf(7)*1e9,5)])
    disp(['Z8, y-Coma           = ',num2str(zernf(8)*1e9,5)])
    disp(['Z9, spherical        = ',num2str(zernf(9)*1e9,5)])
    
    % remove piston, tip, tilt
    zCorrected = z - zernf(1:3) * A(:,1:3)';
    zCorrectedMean = mean(zCorrected);
    zCorrectedPV = max(zCorrected) - min(zCorrected);
    zCorrectedRms = norm(zCorrected)/sqrt(length(zCorrected));
    disp(['mean (piston,tip & tilt removed) = ',num2str(zCorrectedMean,5),' (m)'])
    disp(['p-v (piston,tip & tilt removed)  = ',num2str(zCorrectedPV,5),' (m)'])
    disp(['rms (piston,tip & tilt removed)  = ',num2str(zCorrectedRms,5),' (m)'])
    